@@ -19,32 +19,51 @@
 
 import argparse
 import functools
-import importlib
 import re
 import sys
+from types import ModuleType
+from typing import Any, Callable, cast, Dict, Optional, Tuple, Type, TypeVar, Union
+
+from requests.structures import CaseInsensitiveDict
 
 import gitlab.config
+from gitlab.base import RESTObject
 
-camel_re = re.compile("(.)([A-Z])")
+# This regex is based on:
+# https://github.com/jpvanhal/inflection/blob/master/inflection/__init__.py
+camel_upperlower_regex = re.compile(r"([A-Z]+)([A-Z][a-z])")
+camel_lowerupper_regex = re.compile(r"([a-z\d])([A-Z])")
 
 # custom_actions = {
 #    cls: {
 #        action: (mandatory_args, optional_args, in_obj),
 #    },
 # }
-custom_actions = {}
+custom_actions: Dict[str, Dict[str, Tuple[Tuple[str, ...], Tuple[str, ...], bool]]] = {}
 
 
-def register_custom_action(cls_names, mandatory=tuple(), optional=tuple()):
-    def wrap(f):
+# For an explanation of how these type-hints work see:
+# https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
+#
+# The goal here is that functions which get decorated will retain their types.
+__F = TypeVar("__F", bound=Callable[..., Any])
+
+
+def register_custom_action(
+    cls_names: Union[str, Tuple[str, ...]],
+    mandatory: Tuple[str, ...] = tuple(),
+    optional: Tuple[str, ...] = tuple(),
+) -> Callable[[__F], __F]:
+    def wrap(f: __F) -> __F:
         @functools.wraps(f)
-        def wrapped_f(*args, **kwargs):
+        def wrapped_f(*args: Any, **kwargs: Any) -> Any:
             return f(*args, **kwargs)
 
         # in_obj defines whether the method belongs to the obj or the manager
         in_obj = True
-        classes = cls_names
-        if type(cls_names) != tuple:
+        if isinstance(cls_names, tuple):
+            classes = cls_names
+        else:
             classes = (cls_names,)
 
         for cls_name in classes:
@@ -58,27 +77,32 @@ def register_custom_action(cls_names, mandatory=tuple(), optional=tuple()):
             action = f.__name__.replace("_", "-")
             custom_actions[final_name][action] = (mandatory, optional, in_obj)
 
-        return wrapped_f
+        return cast(__F, wrapped_f)
 
     return wrap
 
 
-def die(msg, e=None):
+def die(msg: str, e: Optional[Exception] = None) -> None:
     if e:
         msg = "%s (%s)" % (msg, e)
     sys.stderr.write(msg + "\n")
     sys.exit(1)
 
 
-def what_to_cls(what):
-    return "".join([s.capitalize() for s in what.split("-")])
+def what_to_cls(what: str, namespace: ModuleType) -> Type[RESTObject]:
+    classes = CaseInsensitiveDict(namespace.__dict__)
+    lowercase_class = what.replace("-", "")
+
+    return classes[lowercase_class]
 
 
-def cls_to_what(cls):
-    return camel_re.sub(r"\1-\2", cls.__name__).lower()
+def cls_to_what(cls: RESTObject) -> str:
+    dasherized_uppercase = camel_upperlower_regex.sub(r"\1-\2", cls.__name__)
+    dasherized_lowercase = camel_lowerupper_regex.sub(r"\1-\2", dasherized_uppercase)
+    return dasherized_lowercase.lower()
 
 
-def _get_base_parser(add_help=True):
+def _get_base_parser(add_help: bool = True) -> argparse.ArgumentParser:
     parser = argparse.ArgumentParser(
         add_help=add_help, description="GitLab API Command Line Interface"
     )
@@ -130,12 +154,16 @@ def _get_base_parser(add_help=True):
     return parser
 
 
-def _get_parser(cli_module):
+def _get_parser() -> argparse.ArgumentParser:
+    # NOTE: We must delay import of gitlab.v4.cli until now or
+    # otherwise it will cause circular import errors
+    import gitlab.v4.cli
+
     parser = _get_base_parser()
-    return cli_module.extend_parser(parser)
+    return gitlab.v4.cli.extend_parser(parser)
 
 
-def _parse_value(v):
+def _parse_value(v: Any) -> Any:
     if isinstance(v, str) and v.startswith("@"):
         # If the user-provided value starts with @, we try to read the file
         # path provided after @ as the real value. Exit on any error.
@@ -149,7 +177,18 @@ def _parse_value(v):
     return v
 
 
-def main():
+def docs() -> argparse.ArgumentParser:
+    """
+    Provide a statically generated parser for sphinx only, so we don't need
+    to provide dummy gitlab config for readthedocs.
+    """
+    if "sphinx" not in sys.modules:
+        sys.exit("Docs parser is only intended for build_sphinx")
+
+    return _get_parser()
+
+
+def main() -> None:
     if "--version" in sys.argv:
         print(gitlab.__version__)
         sys.exit(0)
@@ -159,7 +198,7 @@ def main():
     # This first parsing step is used to find the gitlab config to use, and
     # load the propermodule (v3 or v4) accordingly. At that point we don't have
     # any subparser setup
-    (options, args) = parser.parse_known_args(sys.argv)
+    (options, _) = parser.parse_known_args(sys.argv)
     try:
         config = gitlab.config.GitlabConfigParser(options.gitlab, options.config_file)
     except gitlab.config.ConfigError as e:
@@ -167,17 +206,19 @@ def main():
             parser.print_help()
             sys.exit(0)
         sys.exit(e)
-    cli_module = importlib.import_module("gitlab.v%s.cli" % config.api_version)
+    # We only support v4 API at this time
+    if config.api_version not in ("4",):
+        raise ModuleNotFoundError(name="gitlab.v%s.cli" % config.api_version)
 
     # Now we build the entire set of subcommands and do the complete parsing
-    parser = _get_parser(cli_module)
+    parser = _get_parser()
     try:
-        import argcomplete
+        import argcomplete  # type: ignore
 
         argcomplete.autocomplete(parser)
     except Exception:
         pass
-    args = parser.parse_args(sys.argv[1:])
+    args = parser.parse_args()
 
     config_files = args.config_file
     gitlab_id = args.gitlab
@@ -190,7 +231,7 @@ def main():
     action = args.whaction
     what = args.what
 
-    args = args.__dict__
+    args_dict = vars(args)
     # Remove CLI behavior-related args
     for item in (
         "gitlab",
@@ -202,8 +243,8 @@ def main():
         "version",
         "output",
     ):
-        args.pop(item)
-    args = {k: _parse_value(v) for k, v in args.items() if v is not None}
+        args_dict.pop(item)
+    args_dict = {k: _parse_value(v) for k, v in args_dict.items() if v is not None}
 
     try:
         gl = gitlab.Gitlab.from_config(gitlab_id, config_files)
@@ -215,6 +256,4 @@ def main():
     if debug:
         gl.enable_debug()
 
-    cli_module.run(gl, what, action, args, verbose, output, fields)
-
-    sys.exit(0)
+    gitlab.v4.cli.run(gl, what, action, args_dict, verbose, output, fields)
