@@ -4,8 +4,9 @@ GitLab API: https://docs.gitlab.com/ee/api/repositories.html
 Currently this module only contains repository-related methods for projects.
 """
 
-from gitlab import cli, utils
+from gitlab import cli
 from gitlab import exceptions as exc
+from gitlab import utils
 
 
 class RepositoryMixin:
